@@ -4,12 +4,12 @@
  *
  *	  Routines for opclass manipulation commands
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/commands/opclasscmds.c,v 1.6 2002/10/04 22:19:29 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/commands/opclasscmds.c,v 1.21 2003/09/26 15:27:31 petere Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -78,15 +78,18 @@ DefineOpClass(CreateOpClassStmt *stmt)
 	/* Check we have creation rights in target namespace */
 	aclresult = pg_namespace_aclcheck(namespaceoid, GetUserId(), ACL_CREATE);
 	if (aclresult != ACLCHECK_OK)
-		aclcheck_error(aclresult, get_namespace_name(namespaceoid));
+		aclcheck_error(aclresult, ACL_KIND_NAMESPACE,
+					   get_namespace_name(namespaceoid));
 
 	/* Get necessary info about access method */
 	tup = SearchSysCache(AMNAME,
 						 CStringGetDatum(stmt->amname),
 						 0, 0, 0);
 	if (!HeapTupleIsValid(tup))
-		elog(ERROR, "DefineOpClass: access method \"%s\" not found",
-			 stmt->amname);
+		ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_OBJECT),
+				 errmsg("access method \"%s\" does not exist",
+						stmt->amname)));
 
 	amoid = HeapTupleGetOid(tup);
 	numOperators = ((Form_pg_am) GETSTRUCT(tup))->amstrategies;
@@ -100,11 +103,13 @@ DefineOpClass(CreateOpClassStmt *stmt)
 	 * Currently, we require superuser privileges to create an opclass.
 	 * This seems necessary because we have no way to validate that the
 	 * offered set of operators and functions are consistent with the AM's
-	 * expectations.  It would be nice to provide such a check someday,
-	 * if it can be done without solving the halting problem :-(
+	 * expectations.  It would be nice to provide such a check someday, if
+	 * it can be done without solving the halting problem :-(
 	 */
 	if (!superuser())
-		elog(ERROR, "Must be superuser to create an operator class");
+		ereport(ERROR,
+				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+			   errmsg("must be superuser to create an operator class")));
 
 	/* Look up the datatype */
 	typeoid = typenameTypeId(stmt->datatype);
@@ -113,7 +118,8 @@ DefineOpClass(CreateOpClassStmt *stmt)
 	/* XXX this is unnecessary given the superuser check above */
 	/* Check we have ownership of the datatype */
 	if (!pg_type_ownercheck(typeoid, GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER, format_type_be(typeoid));
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_TYPE,
+					   format_type_be(typeoid));
 #endif
 
 	/* Storage datatype is optional */
@@ -124,12 +130,9 @@ DefineOpClass(CreateOpClassStmt *stmt)
 	 * do this mainly so that we can detect duplicate strategy numbers and
 	 * support-proc numbers.
 	 */
-	operators = (Oid *) palloc(sizeof(Oid) * numOperators);
-	MemSet(operators, 0, sizeof(Oid) * numOperators);
-	procedures = (Oid *) palloc(sizeof(Oid) * numProcs);
-	MemSet(procedures, 0, sizeof(Oid) * numProcs);
-	recheck = (bool *) palloc(sizeof(bool) * numOperators);
-	MemSet(recheck, 0, sizeof(bool) * numOperators);
+	operators = (Oid *) palloc0(sizeof(Oid) * numOperators);
+	procedures = (Oid *) palloc0(sizeof(Oid) * numProcs);
+	recheck = (bool *) palloc0(sizeof(bool) * numOperators);
 
 	/*
 	 * Scan the "items" list to obtain additional info.
@@ -146,66 +149,73 @@ DefineOpClass(CreateOpClassStmt *stmt)
 		{
 			case OPCLASS_ITEM_OPERATOR:
 				if (item->number <= 0 || item->number > numOperators)
-					elog(ERROR, "DefineOpClass: invalid operator number %d,"
-						 " must be between 1 and %d",
-						 item->number, numOperators);
+					ereport(ERROR,
+							(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+							 errmsg("invalid operator number %d,"
+									" must be between 1 and %d",
+									item->number, numOperators)));
 				if (operators[item->number - 1] != InvalidOid)
-					elog(ERROR, "DefineOpClass: operator number %d appears more than once",
-						 item->number);
+					ereport(ERROR,
+							(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+					  errmsg("operator number %d appears more than once",
+							 item->number)));
 				if (item->args != NIL)
 				{
 					TypeName   *typeName1 = (TypeName *) lfirst(item->args);
 					TypeName   *typeName2 = (TypeName *) lsecond(item->args);
 
 					operOid = LookupOperNameTypeNames(item->name,
-													typeName1, typeName2,
-													  "DefineOpClass");
-					/* No need to check for error */
+													  typeName1,
+													  typeName2,
+													  false);
 				}
 				else
 				{
 					/* Default to binary op on input datatype */
-					operOid = LookupOperName(item->name, typeoid, typeoid);
-					if (!OidIsValid(operOid))
-						elog(ERROR, "DefineOpClass: Operator '%s' for types '%s' and '%s' does not exist",
-							 NameListToString(item->name),
-							 format_type_be(typeoid),
-							 format_type_be(typeoid));
+					operOid = LookupOperName(item->name, typeoid, typeoid,
+											 false);
 				}
 				/* Caller must have execute permission on operators */
 				funcOid = get_opcode(operOid);
 				aclresult = pg_proc_aclcheck(funcOid, GetUserId(),
 											 ACL_EXECUTE);
 				if (aclresult != ACLCHECK_OK)
-					aclcheck_error(aclresult, get_func_name(funcOid));
+					aclcheck_error(aclresult, ACL_KIND_PROC,
+								   get_func_name(funcOid));
 				operators[item->number - 1] = operOid;
 				recheck[item->number - 1] = item->recheck;
 				break;
 			case OPCLASS_ITEM_FUNCTION:
 				if (item->number <= 0 || item->number > numProcs)
-					elog(ERROR, "DefineOpClass: invalid procedure number %d,"
-						 " must be between 1 and %d",
-						 item->number, numProcs);
+					ereport(ERROR,
+							(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+							 errmsg("invalid procedure number %d,"
+									" must be between 1 and %d",
+									item->number, numProcs)));
 				if (procedures[item->number - 1] != InvalidOid)
-					elog(ERROR, "DefineOpClass: procedure number %d appears more than once",
-						 item->number);
+					ereport(ERROR,
+							(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+							 errmsg("procedure number %d appears more than once",
+									item->number)));
 				funcOid = LookupFuncNameTypeNames(item->name, item->args,
-												  "DefineOpClass");
+												  false);
 				/* Caller must have execute permission on functions */
 				aclresult = pg_proc_aclcheck(funcOid, GetUserId(),
 											 ACL_EXECUTE);
 				if (aclresult != ACLCHECK_OK)
-					aclcheck_error(aclresult, get_func_name(funcOid));
+					aclcheck_error(aclresult, ACL_KIND_PROC,
+								   get_func_name(funcOid));
 				procedures[item->number - 1] = funcOid;
 				break;
 			case OPCLASS_ITEM_STORAGETYPE:
 				if (OidIsValid(storageoid))
-					elog(ERROR, "DefineOpClass: storage type specified more than once");
+					ereport(ERROR,
+							(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+					   errmsg("storage type specified more than once")));
 				storageoid = typenameTypeId(item->storedtype);
 				break;
 			default:
-				elog(ERROR, "DefineOpClass: bogus item type %d",
-					 item->itemtype);
+				elog(ERROR, "unrecognized item type: %d", item->itemtype);
 				break;
 		}
 	}
@@ -226,8 +236,10 @@ DefineOpClass(CreateOpClassStmt *stmt)
 			 * favor of adding another boolean column to pg_am ...
 			 */
 			if (amoid != GIST_AM_OID)
-				elog(ERROR, "Storage type may not be different from datatype for access method %s",
-					 stmt->amname);
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+						 errmsg("storage type may not be different from data type for access method \"%s\"",
+								stmt->amname)));
 		}
 	}
 
@@ -242,12 +254,16 @@ DefineOpClass(CreateOpClassStmt *stmt)
 							 CStringGetDatum(opcname),
 							 ObjectIdGetDatum(namespaceoid),
 							 0))
-		elog(ERROR, "Operator class \"%s\" already exists for access method \"%s\"",
-			 opcname, stmt->amname);
+		ereport(ERROR,
+				(errcode(ERRCODE_DUPLICATE_OBJECT),
+				 errmsg("operator class \"%s\" for access method \"%s\" already exists",
+						opcname, stmt->amname)));
 
 	/*
 	 * If we are creating a default opclass, check there isn't one
-	 * already. (XXX should we restrict this test to visible opclasses?)
+	 * already.  (Note we do not restrict this test to visible opclasses;
+	 * this ensures that typcache.c can find unique solutions to its
+	 * questions.)
 	 */
 	if (stmt->isDefault)
 	{
@@ -266,11 +282,13 @@ DefineOpClass(CreateOpClassStmt *stmt)
 			Form_pg_opclass opclass = (Form_pg_opclass) GETSTRUCT(tup);
 
 			if (opclass->opcintype == typeoid && opclass->opcdefault)
-				elog(ERROR, "Can't add class \"%s\" as default for type %s"
-					 "\n\tclass \"%s\" already is the default",
-					 opcname,
-					 TypeNameToString(stmt->datatype),
-					 NameStr(opclass->opcname));
+				ereport(ERROR,
+						(errcode(ERRCODE_DUPLICATE_OBJECT),
+						 errmsg("could not make operator class \"%s\" be default for type %s",
+								opcname,
+								TypeNameToString(stmt->datatype)),
+						 errdetail("Operator class \"%s\" already is the default.",
+								   NameStr(opclass->opcname))));
 		}
 
 		systable_endscan(scan);
@@ -474,8 +492,10 @@ RemoveOpClass(RemoveOpClassStmt *stmt)
 						  CStringGetDatum(stmt->amname),
 						  0, 0, 0);
 	if (!OidIsValid(amID))
-		elog(ERROR, "RemoveOpClass: access method \"%s\" not found",
-			 stmt->amname);
+		ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_OBJECT),
+				 errmsg("access method \"%s\" does not exist",
+						stmt->amname)));
 
 	/*
 	 * Look up the opclass.
@@ -501,16 +521,20 @@ RemoveOpClass(RemoveOpClassStmt *stmt)
 		/* Unqualified opclass name, so search the search path */
 		opcID = OpclassnameGetOpcid(amID, opcname);
 		if (!OidIsValid(opcID))
-			elog(ERROR, "RemoveOpClass: operator class \"%s\" not supported by access method \"%s\"",
-				 opcname, stmt->amname);
+			ereport(ERROR,
+					(errcode(ERRCODE_UNDEFINED_OBJECT),
+					 errmsg("operator class \"%s\" does not exist for access method \"%s\"",
+							opcname, stmt->amname)));
 		tuple = SearchSysCache(CLAOID,
 							   ObjectIdGetDatum(opcID),
 							   0, 0, 0);
 	}
 
 	if (!HeapTupleIsValid(tuple))
-		elog(ERROR, "RemoveOpClass: operator class \"%s\" not supported by access method \"%s\"",
-			 NameListToString(stmt->opclassname), stmt->amname);
+		ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_OBJECT),
+				 errmsg("operator class \"%s\" does not exist for access method \"%s\"",
+					NameListToString(stmt->opclassname), stmt->amname)));
 
 	opcID = HeapTupleGetOid(tuple);
 
@@ -518,7 +542,7 @@ RemoveOpClass(RemoveOpClassStmt *stmt)
 	if (!pg_opclass_ownercheck(opcID, GetUserId()) &&
 		!pg_namespace_ownercheck(((Form_pg_opclass) GETSTRUCT(tuple))->opcnamespace,
 								 GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER,
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_OPCLASS,
 					   NameListToString(stmt->opclassname));
 
 	ReleaseSysCache(tuple);
@@ -553,8 +577,7 @@ RemoveOpClassById(Oid opclassOid)
 						 ObjectIdGetDatum(opclassOid),
 						 0, 0, 0);
 	if (!HeapTupleIsValid(tup)) /* should not happen */
-		elog(ERROR, "RemoveOpClassById: couldn't find pg_class entry %u",
-			 opclassOid);
+		elog(ERROR, "cache lookup failed for opclass %u", opclassOid);
 
 	simple_heap_delete(rel, &tup->t_self);
 
@@ -598,3 +621,104 @@ RemoveOpClassById(Oid opclassOid)
 	systable_endscan(scan);
 	heap_close(rel, RowExclusiveLock);
 }
+
+
+/*
+ * Rename opclass
+ */
+void
+RenameOpClass(List *name, const char *access_method, const char *newname)
+{
+	Oid			opcOid;
+	Oid			amOid;
+	Oid			namespaceOid;
+	char	   *schemaname;
+	char	   *opcname;
+	HeapTuple	tup;
+	Relation	rel;
+	AclResult	aclresult;
+
+	amOid = GetSysCacheOid(AMNAME,
+						   CStringGetDatum(access_method),
+						   0, 0, 0);
+	if (!OidIsValid(amOid))
+		ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_OBJECT),
+				 errmsg("access method \"%s\" does not exist",
+						access_method)));
+
+	rel = heap_openr(OperatorClassRelationName, RowExclusiveLock);
+
+	/*
+	 * Look up the opclass
+	 */
+	DeconstructQualifiedName(name, &schemaname, &opcname);
+
+	if (schemaname)
+	{
+		namespaceOid = LookupExplicitNamespace(schemaname);
+
+		tup = SearchSysCacheCopy(CLAAMNAMENSP,
+								 ObjectIdGetDatum(amOid),
+								 PointerGetDatum(opcname),
+								 ObjectIdGetDatum(namespaceOid),
+								 0);
+		if (!HeapTupleIsValid(tup))
+			ereport(ERROR,
+					(errcode(ERRCODE_UNDEFINED_OBJECT),
+					 errmsg("operator class \"%s\" does not exist for access method \"%s\"",
+							opcname, access_method)));
+
+		opcOid = HeapTupleGetOid(tup);
+	}
+	else
+	{
+		opcOid = OpclassnameGetOpcid(amOid, opcname);
+		if (!OidIsValid(opcOid))
+			ereport(ERROR,
+					(errcode(ERRCODE_UNDEFINED_OBJECT),
+					 errmsg("operator class \"%s\" does not exist for access method \"%s\"",
+							opcname, access_method)));
+
+		tup = SearchSysCacheCopy(CLAOID,
+								 ObjectIdGetDatum(opcOid),
+								 0, 0, 0);
+		if (!HeapTupleIsValid(tup))		/* should not happen */
+			elog(ERROR, "cache lookup failed for opclass %u", opcOid);
+
+		namespaceOid = ((Form_pg_opclass) GETSTRUCT(tup))->opcnamespace;
+	}
+
+	/* make sure the new name doesn't exist */
+	if (SearchSysCacheExists(CLAAMNAMENSP,
+							 ObjectIdGetDatum(amOid),
+							 CStringGetDatum(newname),
+							 ObjectIdGetDatum(namespaceOid),
+							 0))
+	{
+		ereport(ERROR,
+				(errcode(ERRCODE_DUPLICATE_OBJECT),
+				 errmsg("operator class \"%s\" for access method \"%s\" already exists in schema \"%s\"",
+						newname, access_method,
+						get_namespace_name(namespaceOid))));
+	}
+
+	/* must be owner */
+	if (!pg_opclass_ownercheck(opcOid, GetUserId()))
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_OPCLASS,
+					   NameListToString(name));
+
+	/* must have CREATE privilege on namespace */
+	aclresult = pg_namespace_aclcheck(namespaceOid, GetUserId(), ACL_CREATE);
+	if (aclresult != ACLCHECK_OK)
+		aclcheck_error(aclresult, ACL_KIND_NAMESPACE,
+					   get_namespace_name(namespaceOid));
+
+	/* rename */
+	namestrcpy(&(((Form_pg_opclass) GETSTRUCT(tup))->opcname), newname);
+	simple_heap_update(rel, &tup->t_self, tup);
+	CatalogUpdateIndexes(rel, tup);
+
+	heap_close(rel, NoLock);
+	heap_freetuple(tup);
+}
