@@ -3,12 +3,12 @@
  * parse_relation.c
  *	  parser support routines dealing with relations
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/parser/parse_relation.c,v 1.79.2.1 2005/05/29 17:11:07 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/parser/parse_relation.c,v 1.90.2.1 2004/04/18 18:13:31 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -32,6 +32,8 @@
 #include "utils/lsyscache.h"
 #include "utils/syscache.h"
 
+/* GUC parameter */
+bool		add_missing_from;
 
 static Node *scanNameSpaceForRefname(ParseState *pstate, Node *nsnode,
 						const char *refname);
@@ -60,7 +62,8 @@ static void warnAutoRange(ParseState *pstate, RangeVar *relation);
  * An unqualified refname (schemaname == NULL) can match any RTE with matching
  * alias, or matching unqualified relname in the case of alias-less relation
  * RTEs.  It is possible that such a refname matches multiple RTEs in the
- * nearest nesting level that has a match; if so, we report an error via elog.
+ * nearest nesting level that has a match; if so, we report an error via
+ * ereport().
  *
  * A qualified refname (schemaname != NULL) can only match a relation RTE
  * that (a) has no alias and (b) is for the same relation identified by
@@ -166,7 +169,10 @@ scanNameSpaceForRefname(ParseState *pstate, Node *nsnode,
 		if (!result)
 			result = newresult;
 		else if (newresult)
-			elog(ERROR, "Table reference \"%s\" is ambiguous", refname);
+			ereport(ERROR,
+					(errcode(ERRCODE_AMBIGUOUS_ALIAS),
+					 errmsg("table reference \"%s\" is ambiguous",
+							refname)));
 	}
 	else if (IsA(nsnode, List))
 	{
@@ -178,12 +184,14 @@ scanNameSpaceForRefname(ParseState *pstate, Node *nsnode,
 			if (!result)
 				result = newresult;
 			else if (newresult)
-				elog(ERROR, "Table reference \"%s\" is ambiguous", refname);
+				ereport(ERROR,
+						(errcode(ERRCODE_AMBIGUOUS_ALIAS),
+						 errmsg("table reference \"%s\" is ambiguous",
+								refname)));
 		}
 	}
 	else
-		elog(ERROR, "scanNameSpaceForRefname: unexpected node type %d",
-			 nodeTag(nsnode));
+		elog(ERROR, "unrecognized node type: %d", (int) nodeTag(nsnode));
 	return result;
 }
 
@@ -236,7 +244,10 @@ scanNameSpaceForRelid(ParseState *pstate, Node *nsnode, Oid relid)
 		if (!result)
 			result = newresult;
 		else if (newresult)
-			elog(ERROR, "Table reference %u is ambiguous", relid);
+			ereport(ERROR,
+					(errcode(ERRCODE_AMBIGUOUS_ALIAS),
+					 errmsg("table reference %u is ambiguous",
+							relid)));
 	}
 	else if (IsA(nsnode, List))
 	{
@@ -248,12 +259,14 @@ scanNameSpaceForRelid(ParseState *pstate, Node *nsnode, Oid relid)
 			if (!result)
 				result = newresult;
 			else if (newresult)
-				elog(ERROR, "Table reference %u is ambiguous", relid);
+				ereport(ERROR,
+						(errcode(ERRCODE_AMBIGUOUS_ALIAS),
+						 errmsg("table reference %u is ambiguous",
+								relid)));
 		}
 	}
 	else
-		elog(ERROR, "scanNameSpaceForRelid: unexpected node type %d",
-			 nodeTag(nsnode));
+		elog(ERROR, "unrecognized node type: %d", (int) nodeTag(nsnode));
 	return result;
 }
 
@@ -316,8 +329,7 @@ checkNameSpaceConflicts(ParseState *pstate, Node *namespace1,
 			checkNameSpaceConflicts(pstate, lfirst(l), namespace2);
 	}
 	else
-		elog(ERROR, "checkNameSpaceConflicts: unexpected node type %d",
-			 nodeTag(namespace1));
+		elog(ERROR, "unrecognized node type: %d", (int) nodeTag(namespace1));
 }
 
 /*
@@ -339,8 +351,10 @@ scanNameSpaceForConflict(ParseState *pstate, Node *nsnode,
 		if (rte->rtekind == RTE_RELATION && rte->alias == NULL &&
 			rte1 != NULL && rte->relid != rte1->relid)
 			return;				/* no conflict per SQL92 rule */
-		elog(ERROR, "Table name \"%s\" specified more than once",
-			 aliasname1);
+		ereport(ERROR,
+				(errcode(ERRCODE_DUPLICATE_ALIAS),
+				 errmsg("table name \"%s\" specified more than once",
+						aliasname1)));
 	}
 	else if (IsA(nsnode, JoinExpr))
 	{
@@ -349,8 +363,10 @@ scanNameSpaceForConflict(ParseState *pstate, Node *nsnode,
 		if (j->alias)
 		{
 			if (strcmp(j->alias->aliasname, aliasname1) == 0)
-				elog(ERROR, "Table name \"%s\" specified more than once",
-					 aliasname1);
+				ereport(ERROR,
+						(errcode(ERRCODE_DUPLICATE_ALIAS),
+					 errmsg("table name \"%s\" specified more than once",
+							aliasname1)));
 
 			/*
 			 * Tables within an aliased join are invisible from outside
@@ -370,8 +386,7 @@ scanNameSpaceForConflict(ParseState *pstate, Node *nsnode,
 			scanNameSpaceForConflict(pstate, lfirst(l), rte1, aliasname1);
 	}
 	else
-		elog(ERROR, "scanNameSpaceForConflict: unexpected node type %d",
-			 nodeTag(nsnode));
+		elog(ERROR, "unrecognized node type: %d", (int) nodeTag(nsnode));
 }
 
 /*
@@ -405,7 +420,7 @@ RTERangeTablePosn(ParseState *pstate, RangeTblEntry *rte, int *sublevels_up)
 			break;
 	}
 
-	elog(ERROR, "RTERangeTablePosn: RTE not found (internal error)");
+	elog(ERROR, "RTE not found (internal error)");
 	return 0;					/* keep compiler quiet */
 }
 
@@ -457,7 +472,10 @@ scanRTEForColumn(ParseState *pstate, RangeTblEntry *rte, char *colname)
 				get_rte_attribute_is_dropped(rte, attnum))
 				continue;
 			if (result)
-				elog(ERROR, "Column reference \"%s\" is ambiguous", colname);
+				ereport(ERROR,
+						(errcode(ERRCODE_AMBIGUOUS_COLUMN),
+						 errmsg("column reference \"%s\" is ambiguous",
+								colname)));
 			result = (Node *) make_var(pstate, rte, attnum);
 			rte->checkForRead = true;
 		}
@@ -495,13 +513,14 @@ scanRTEForColumn(ParseState *pstate, RangeTblEntry *rte, char *colname)
 }
 
 /*
- * colnameToVar
+ * colNameToVar
  *	  Search for an unqualified column name.
  *	  If found, return the appropriate Var node (or expression).
  *	  If not found, return NULL.  If the name proves ambiguous, raise error.
+ *	  If localonly is true, only names in the innermost query are considered.
  */
 Node *
-colnameToVar(ParseState *pstate, char *colname)
+colNameToVar(ParseState *pstate, char *colname, bool localonly)
 {
 	Node	   *result = NULL;
 	ParseState *orig_pstate = pstate;
@@ -544,20 +563,22 @@ colnameToVar(ParseState *pstate, char *colname)
 				newresult = scanRTEForColumn(orig_pstate, rte, colname);
 			}
 			else
-				elog(ERROR, "colnameToVar: unexpected node type %d",
-					 nodeTag(nsnode));
+				elog(ERROR, "unrecognized node type: %d",
+					 (int) nodeTag(nsnode));
 
 			if (newresult)
 			{
 				if (result)
-					elog(ERROR, "Column reference \"%s\" is ambiguous",
-						 colname);
+					ereport(ERROR,
+							(errcode(ERRCODE_AMBIGUOUS_COLUMN),
+						   errmsg("column reference \"%s\" is ambiguous",
+								  colname)));
 				result = newresult;
 			}
 		}
 
-		if (result != NULL)
-			break;				/* found */
+		if (result != NULL || localonly)
+			break;				/* found, or don't want to look at parent */
 
 		pstate = pstate->parentParseState;
 		levels_up++;
@@ -643,8 +664,10 @@ addRangeTableEntry(ParseState *pstate,
 	 */
 	maxattrs = RelationGetNumberOfAttributes(rel);
 	if (maxattrs < numaliases)
-		elog(ERROR, "Table \"%s\" has %d columns available but %d columns specified",
-			 RelationGetRelationName(rel), maxattrs, numaliases);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
+				 errmsg("table \"%s\" has %d columns available but %d columns specified",
+				   RelationGetRelationName(rel), maxattrs, numaliases)));
 
 	/* fill in any unspecified alias columns using actual column names */
 	for (varattno = numaliases; varattno < maxattrs; varattno++)
@@ -736,8 +759,10 @@ addRangeTableEntryForRelation(ParseState *pstate,
 	 */
 	maxattrs = RelationGetNumberOfAttributes(rel);
 	if (maxattrs < numaliases)
-		elog(ERROR, "Table \"%s\" has %d columns available but %d columns specified",
-			 RelationGetRelationName(rel), maxattrs, numaliases);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
+				 errmsg("table \"%s\" has %d columns available but %d columns specified",
+				   RelationGetRelationName(rel), maxattrs, numaliases)));
 
 	/* fill in any unspecified alias columns using actual column names */
 	for (varattno = numaliases; varattno < maxattrs; varattno++)
@@ -829,8 +854,10 @@ addRangeTableEntryForSubquery(ParseState *pstate,
 		}
 	}
 	if (varattno < numaliases)
-		elog(ERROR, "Table \"%s\" has %d columns available but %d columns specified",
-			 refname, varattno, numaliases);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
+				 errmsg("table \"%s\" has %d columns available but %d columns specified",
+						refname, varattno, numaliases)));
 
 	rte->eref = eref;
 
@@ -904,7 +931,9 @@ addRangeTableEntryForFunction(ParseState *pstate,
 		 * pseudo-type
 		 */
 		if (funcrettype != RECORDOID)
-			elog(ERROR, "A column definition list is only allowed for functions returning RECORD");
+			ereport(ERROR,
+					(errcode(ERRCODE_SYNTAX_ERROR),
+					 errmsg("a column definition list is only allowed for functions returning \"record\"")));
 	}
 	else
 	{
@@ -913,7 +942,9 @@ addRangeTableEntryForFunction(ParseState *pstate,
 		 * RECORD pseudo-type
 		 */
 		if (funcrettype == RECORDOID)
-			elog(ERROR, "A column definition list is required for functions returning RECORD");
+			ereport(ERROR,
+					(errcode(ERRCODE_SYNTAX_ERROR),
+					 errmsg("a column definition list is required for functions returning \"record\"")));
 	}
 
 	functyptype = get_typtype(funcrettype);
@@ -927,9 +958,9 @@ addRangeTableEntryForFunction(ParseState *pstate,
 		Relation	rel;
 		int			maxattrs;
 
-		if (!OidIsValid(funcrelid))
-			elog(ERROR, "Invalid typrelid for complex type %u",
-				 funcrettype);
+		if (!OidIsValid(funcrelid))		/* shouldn't happen if typtype is
+										 * 'c' */
+			elog(ERROR, "invalid typrelid for complex type %u", funcrettype);
 
 		/*
 		 * Get the rel's relcache entry.  This access ensures that we have
@@ -943,8 +974,11 @@ addRangeTableEntryForFunction(ParseState *pstate,
 		 */
 		maxattrs = RelationGetNumberOfAttributes(rel);
 		if (maxattrs < numaliases)
-			elog(ERROR, "Table \"%s\" has %d columns available but %d columns specified",
-				 RelationGetRelationName(rel), maxattrs, numaliases);
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
+					 errmsg("table \"%s\" has %d columns available but %d columns specified",
+							RelationGetRelationName(rel),
+							maxattrs, numaliases)));
 
 		/* fill in alias columns using actual column names */
 		for (varattno = numaliases; varattno < maxattrs; varattno++)
@@ -969,8 +1003,10 @@ addRangeTableEntryForFunction(ParseState *pstate,
 		 * column named for the function.
 		 */
 		if (numaliases > 1)
-			elog(ERROR, "Too many column aliases specified for function %s",
-				 funcname);
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_COLUMN_REFERENCE),
+			  errmsg("too many column aliases specified for function %s",
+					 funcname)));
 		if (numaliases == 0)
 			eref->colnames = makeList1(makeString(eref->aliasname));
 	}
@@ -990,8 +1026,10 @@ addRangeTableEntryForFunction(ParseState *pstate,
 		}
 	}
 	else
-		elog(ERROR, "Unknown kind of return type specified for function %s",
-			 funcname);
+		ereport(ERROR,
+				(errcode(ERRCODE_DATATYPE_MISMATCH),
+			errmsg("function \"%s\" in FROM has unsupported return type",
+				   funcname)));
 
 	/*----------
 	 * Flags:
@@ -1282,8 +1320,8 @@ expandRTE(ParseState *pstate, RangeTblEntry *rte,
 					int			maxattrs;
 					int			numaliases;
 
-					if (!OidIsValid(funcrelid))
-						elog(ERROR, "Invalid typrelid for complex type %u",
+					if (!OidIsValid(funcrelid)) /* shouldn't happen */
+						elog(ERROR, "invalid typrelid for complex type %u",
 							 funcrettype);
 
 					rel = relation_open(funcrelid, AccessShareLock);
@@ -1372,7 +1410,7 @@ expandRTE(ParseState *pstate, RangeTblEntry *rte,
 							varnode = makeVar(rtindex,
 											  attnum,
 											  atttypid,
-											  colDef->typename->typmod,
+											  -1,
 											  sublevels_up);
 
 							*colvars = lappend(*colvars, varnode);
@@ -1380,7 +1418,9 @@ expandRTE(ParseState *pstate, RangeTblEntry *rte,
 					}
 				}
 				else
-					elog(ERROR, "Unknown kind of return type specified for function");
+					ereport(ERROR,
+							(errcode(ERRCODE_DATATYPE_MISMATCH),
+							 errmsg("function in FROM has unsupported return type")));
 			}
 			break;
 		case RTE_JOIN:
@@ -1422,8 +1462,7 @@ expandRTE(ParseState *pstate, RangeTblEntry *rte,
 			}
 			break;
 		default:
-			elog(ERROR, "expandRTE: unsupported RTE kind %d",
-				 (int) rte->rtekind);
+			elog(ERROR, "unrecognized RTE kind: %d", (int) rte->rtekind);
 	}
 }
 
@@ -1447,12 +1486,12 @@ expandRelAttrs(ParseState *pstate, RangeTblEntry *rte)
 		Node	   *varnode = (Node *) lfirst(vars);
 		TargetEntry *te = makeNode(TargetEntry);
 
-		te->resdom = makeResdom((AttrNumber) (pstate->p_last_resno)++,
+		te->resdom = makeResdom((AttrNumber) pstate->p_next_resno++,
 								exprType(varnode),
 								exprTypmod(varnode),
 								label,
 								false);
-		te->expr = varnode;
+		te->expr = (Expr *) varnode;
 		te_list = lappend(te_list, te);
 
 		names = lnext(names);
@@ -1478,8 +1517,6 @@ expandRelAttrs(ParseState *pstate, RangeTblEntry *rte)
 char *
 get_rte_attribute_name(RangeTblEntry *rte, AttrNumber attnum)
 {
-	char	   *attname;
-
 	if (attnum == InvalidAttrNumber)
 		return "*";
 
@@ -1497,13 +1534,7 @@ get_rte_attribute_name(RangeTblEntry *rte, AttrNumber attnum)
 	 * built (which can easily happen for rules).
 	 */
 	if (rte->rtekind == RTE_RELATION)
-	{
-		attname = get_attname(rte->relid, attnum);
-		if (attname == NULL)
-			elog(ERROR, "cache lookup of attribute %d in relation %u failed",
-				 attnum, rte->relid);
-		return attname;
-	}
+		return get_relid_attribute_name(rte->relid, attnum);
 
 	/*
 	 * Otherwise use the column name from eref.  There should always be
@@ -1512,7 +1543,8 @@ get_rte_attribute_name(RangeTblEntry *rte, AttrNumber attnum)
 	if (attnum > 0 && attnum <= length(rte->eref->colnames))
 		return strVal(nth(attnum - 1, rte->eref->colnames));
 
-	elog(ERROR, "Invalid attnum %d for rangetable entry %s",
+	/* else caller gave us a bogus attnum */
+	elog(ERROR, "invalid attnum %d for rangetable entry %s",
 		 attnum, rte->eref->aliasname);
 	return NULL;				/* keep compiler quiet */
 }
@@ -1537,10 +1569,9 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
 									ObjectIdGetDatum(rte->relid),
 									Int16GetDatum(attnum),
 									0, 0);
-				/* this shouldn't happen... */
-				if (!HeapTupleIsValid(tp))
-					elog(ERROR, "Relation \"%s\" does not have attribute %d",
-						 get_rel_name(rte->relid), attnum);
+				if (!HeapTupleIsValid(tp))		/* shouldn't happen */
+					elog(ERROR, "cache lookup failed for attribute %d of relation %u",
+						 attnum, rte->relid);
 				att_tup = (Form_pg_attribute) GETSTRUCT(tp);
 
 				/*
@@ -1548,8 +1579,11 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
 				 * in scanRTEForColumn.
 				 */
 				if (att_tup->attisdropped)
-					elog(ERROR, "Relation \"%s\" has no column \"%s\"",
-					get_rel_name(rte->relid), NameStr(att_tup->attname));
+					ereport(ERROR,
+							(errcode(ERRCODE_UNDEFINED_COLUMN),
+							 errmsg("column \"%s\" of relation \"%s\" does not exist",
+									NameStr(att_tup->attname),
+									get_rel_name(rte->relid))));
 				*vartype = att_tup->atttypid;
 				*vartypmod = att_tup->atttypmod;
 				ReleaseSysCache(tp);
@@ -1558,21 +1592,14 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
 		case RTE_SUBQUERY:
 			{
 				/* Subselect RTE --- get type info from subselect's tlist */
-				List	   *tlistitem;
+				TargetEntry *te = get_tle_by_resno(rte->subquery->targetList,
+												   attnum);
 
-				foreach(tlistitem, rte->subquery->targetList)
-				{
-					TargetEntry *te = (TargetEntry *) lfirst(tlistitem);
-
-					if (te->resdom->resjunk || te->resdom->resno != attnum)
-						continue;
-					*vartype = te->resdom->restype;
-					*vartypmod = te->resdom->restypmod;
-					return;
-				}
-				/* falling off end of list shouldn't happen... */
-				elog(ERROR, "Subquery %s does not have attribute %d",
-					 rte->eref->aliasname, attnum);
+				if (te == NULL || te->resdom->resjunk)
+					elog(ERROR, "subquery %s does not have attribute %d",
+						 rte->eref->aliasname, attnum);
+				*vartype = te->resdom->restype;
+				*vartypmod = te->resdom->restypmod;
 			}
 			break;
 		case RTE_FUNCTION:
@@ -1592,18 +1619,17 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
 					HeapTuple	tp;
 					Form_pg_attribute att_tup;
 
-					if (!OidIsValid(funcrelid))
-						elog(ERROR, "Invalid typrelid for complex type %u",
+					if (!OidIsValid(funcrelid)) /* shouldn't happen */
+						elog(ERROR, "invalid typrelid for complex type %u",
 							 funcrettype);
 
 					tp = SearchSysCache(ATTNUM,
 										ObjectIdGetDatum(funcrelid),
 										Int16GetDatum(attnum),
 										0, 0);
-					/* this shouldn't happen... */
-					if (!HeapTupleIsValid(tp))
-						elog(ERROR, "Relation \"%s\" does not have attribute %d",
-							 get_rel_name(funcrelid), attnum);
+					if (!HeapTupleIsValid(tp))	/* shouldn't happen */
+						elog(ERROR, "cache lookup failed for attribute %d of relation %u",
+							 attnum, funcrelid);
 					att_tup = (Form_pg_attribute) GETSTRUCT(tp);
 
 					/*
@@ -1611,9 +1637,11 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
 					 * notes in scanRTEForColumn.
 					 */
 					if (att_tup->attisdropped)
-						elog(ERROR, "Relation \"%s\" has no column \"%s\"",
-							 get_rel_name(funcrelid),
-							 NameStr(att_tup->attname));
+						ereport(ERROR,
+								(errcode(ERRCODE_UNDEFINED_COLUMN),
+								 errmsg("column \"%s\" of relation \"%s\" does not exist",
+										NameStr(att_tup->attname),
+										get_rel_name(funcrelid))));
 					*vartype = att_tup->atttypid;
 					*vartypmod = att_tup->atttypmod;
 					ReleaseSysCache(tp);
@@ -1631,10 +1659,12 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
 					ColumnDef  *colDef = nth(attnum - 1, coldeflist);
 
 					*vartype = typenameTypeId(colDef->typename);
-					*vartypmod = colDef->typename->typmod;
+					*vartypmod = -1;
 				}
 				else
-					elog(ERROR, "Unknown kind of return type specified for function");
+					ereport(ERROR,
+							(errcode(ERRCODE_DATATYPE_MISMATCH),
+							 errmsg("function in FROM has unsupported return type")));
 			}
 			break;
 		case RTE_JOIN:
@@ -1652,8 +1682,7 @@ get_rte_attribute_type(RangeTblEntry *rte, AttrNumber attnum,
 			}
 			break;
 		default:
-			elog(ERROR, "get_rte_attribute_type: unsupported RTE kind %d",
-				 (int) rte->rtekind);
+			elog(ERROR, "unrecognized RTE kind: %d", (int) rte->rtekind);
 	}
 }
 
@@ -1678,10 +1707,9 @@ get_rte_attribute_is_dropped(RangeTblEntry *rte, AttrNumber attnum)
 									ObjectIdGetDatum(rte->relid),
 									Int16GetDatum(attnum),
 									0, 0);
-				/* this shouldn't happen... */
-				if (!HeapTupleIsValid(tp))
-					elog(ERROR, "Relation \"%s\" does not have attribute %d",
-						 get_rel_name(rte->relid), attnum);
+				if (!HeapTupleIsValid(tp))		/* shouldn't happen */
+					elog(ERROR, "cache lookup failed for attribute %d of relation %u",
+						 attnum, rte->relid);
 				att_tup = (Form_pg_attribute) GETSTRUCT(tp);
 				result = att_tup->attisdropped;
 				ReleaseSysCache(tp);
@@ -1711,10 +1739,9 @@ get_rte_attribute_is_dropped(RangeTblEntry *rte, AttrNumber attnum)
 										ObjectIdGetDatum(funcrelid),
 										Int16GetDatum(attnum),
 										0, 0);
-					/* this shouldn't happen... */
-					if (!HeapTupleIsValid(tp))
-						elog(ERROR, "Relation %s does not have attribute %d",
-							 get_rel_name(funcrelid), attnum);
+					if (!HeapTupleIsValid(tp))	/* shouldn't happen */
+						elog(ERROR, "cache lookup failed for attribute %d of relation %u",
+							 attnum, funcrelid);
 					att_tup = (Form_pg_attribute) GETSTRUCT(tp);
 					result = att_tup->attisdropped;
 					ReleaseSysCache(tp);
@@ -1729,14 +1756,36 @@ get_rte_attribute_is_dropped(RangeTblEntry *rte, AttrNumber attnum)
 			}
 			break;
 		default:
-			elog(ERROR, "get_rte_attribute_is_dropped: unsupported RTE kind %d",
-				 (int) rte->rtekind);
+			elog(ERROR, "unrecognized RTE kind: %d", (int) rte->rtekind);
 			result = false;		/* keep compiler quiet */
 	}
 
 	return result;
 }
 
+/*
+ * Given a targetlist and a resno, return the matching TargetEntry
+ *
+ * Returns NULL if resno is not present in list.
+ *
+ * Note: we need to search, rather than just indexing with nth(), because
+ * not all tlists are sorted by resno.
+ */
+TargetEntry *
+get_tle_by_resno(List *tlist, AttrNumber resno)
+{
+	List	   *i;
+
+	foreach(i, tlist)
+	{
+		TargetEntry *tle = (TargetEntry *) lfirst(i);
+
+		if (tle->resdom->resno == resno)
+			return tle;
+	}
+	return NULL;
+}
+
 /*
  *	given relation and att name, return id of variable
  *
@@ -1767,9 +1816,11 @@ attnameAttNum(Relation rd, const char *attname, bool sysColOK)
 	}
 
 	/* on failure */
-	elog(ERROR, "Relation \"%s\" has no column \"%s\"",
-		 RelationGetRelationName(rd), attname);
-	return InvalidAttrNumber;	/* lint */
+	ereport(ERROR,
+			(errcode(ERRCODE_UNDEFINED_COLUMN),
+			 errmsg("column \"%s\" of relation \"%s\" does not exist",
+					attname, RelationGetRelationName(rd))));
+	return InvalidAttrNumber;	/* keep compiler quiet */
 }
 
 /* specialAttNum()
@@ -1812,7 +1863,7 @@ attnumAttName(Relation rd, int attid)
 		return &sysatt->attname;
 	}
 	if (attid > rd->rd_att->natts)
-		elog(ERROR, "attnumAttName: invalid attribute number %d", attid);
+		elog(ERROR, "invalid attribute number %d", attid);
 	return &rd->rd_att->attrs[attid - 1]->attname;
 }
 
@@ -1834,14 +1885,16 @@ attnumTypeId(Relation rd, int attid)
 		return sysatt->atttypid;
 	}
 	if (attid > rd->rd_att->natts)
-		elog(ERROR, "attnumTypeId: invalid attribute number %d", attid);
+		elog(ERROR, "invalid attribute number %d", attid);
 	return rd->rd_att->attrs[attid - 1]->atttypid;
 }
 
 /*
- * Generate a warning about an implicit RTE, if appropriate.
+ * Generate a warning or error about an implicit RTE, if appropriate.
  *
- * Our current theory on this is that we should allow "SELECT foo.*"
+ * If ADD_MISSING_FROM is not enabled, raise an error.
+ *
+ * Our current theory on warnings is that we should allow "SELECT foo.*"
  * but warn about a mixture of explicit and implicit RTEs.
  */
 static void
@@ -1850,6 +1903,20 @@ warnAutoRange(ParseState *pstate, RangeVar *relation)
 	bool		foundInFromCl = false;
 	List	   *temp;
 
+	if (!add_missing_from)
+	{
+		if (pstate->parentParseState != NULL)
+			ereport(ERROR,
+					(errcode(ERRCODE_UNDEFINED_TABLE),
+					 errmsg("missing FROM-clause entry in subquery for table \"%s\"",
+							relation->relname)));
+		else
+			ereport(ERROR,
+					(errcode(ERRCODE_UNDEFINED_TABLE),
+					 errmsg("missing FROM-clause entry for table \"%s\"",
+							relation->relname)));
+	}
+
 	foreach(temp, pstate->p_rtable)
 	{
 		RangeTblEntry *rte = lfirst(temp);
@@ -1861,7 +1928,16 @@ warnAutoRange(ParseState *pstate, RangeVar *relation)
 		}
 	}
 	if (foundInFromCl)
-		elog(NOTICE, "Adding missing FROM-clause entry%s for table \"%s\"",
-			 pstate->parentParseState != NULL ? " in subquery" : "",
-			 relation->relname);
+	{
+		if (pstate->parentParseState != NULL)
+			ereport(NOTICE,
+					(errcode(ERRCODE_UNDEFINED_TABLE),
+					 errmsg("adding missing FROM-clause entry in subquery for table \"%s\"",
+							relation->relname)));
+		else
+			ereport(NOTICE,
+					(errcode(ERRCODE_UNDEFINED_TABLE),
+			  errmsg("adding missing FROM-clause entry for table \"%s\"",
+					 relation->relname)));
+	}
 }
