@@ -5,11 +5,11 @@
  *
  * Code originally contributed by Adriaan Joubert.
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/utils/adt/varbit.c,v 1.26.2.1 2007/08/21 02:40:40 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/utils/adt/varbit.c,v 1.37 2003/09/25 06:58:05 petere Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -17,6 +17,7 @@
 #include "postgres.h"
 
 #include "catalog/pg_type.h"
+#include "libpq/pqformat.h"
 #include "utils/array.h"
 #include "utils/fmgroids.h"
 #include "utils/memutils.h"
@@ -103,13 +104,14 @@ bit_in(PG_FUNCTION_ARGS)
 	if (atttypmod <= 0)
 		atttypmod = bitlen;
 	else if (bitlen != atttypmod)
-		elog(ERROR, "Bit string length %d does not match type BIT(%d)",
-			 bitlen, atttypmod);
+		ereport(ERROR,
+				(errcode(ERRCODE_STRING_DATA_LENGTH_MISMATCH),
+			   errmsg("bit string length %d does not match type bit(%d)",
+					  bitlen, atttypmod)));
 
 	len = VARBITTOTALLEN(atttypmod);
-	result = (VarBit *) palloc(len);
 	/* set to 0 so that *r is always initialised and string is zero-padded */
-	MemSet(result, 0, len);
+	result = (VarBit *) palloc0(len);
 	VARATT_SIZEP(result) = len;
 	VARBITLEN(result) = atttypmod;
 
@@ -124,7 +126,11 @@ bit_in(PG_FUNCTION_ARGS)
 			if (*sp == '1')
 				*r |= x;
 			else if (*sp != '0')
-				elog(ERROR, "Cannot parse '%c' as a binary digit", *sp);
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
+						 errmsg("\"%c\" is not a valid binary digit",
+								*sp)));
+
 			x >>= 1;
 			if (x == 0)
 			{
@@ -145,7 +151,11 @@ bit_in(PG_FUNCTION_ARGS)
 			else if (*sp >= 'a' && *sp <= 'f')
 				x = (bits8) (*sp - 'a') + 10;
 			else
-				elog(ERROR, "Cannot parse '%c' as a hex digit", *sp);
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
+						 errmsg("\"%c\" is not a valid hexadecimal digit",
+								*sp)));
+
 			if (bc)
 			{
 				*r++ |= x;
@@ -205,6 +215,26 @@ bit_out(PG_FUNCTION_ARGS)
 #endif
 }
 
+/*
+ *		bit_recv			- converts external binary format to bit
+ */
+Datum
+bit_recv(PG_FUNCTION_ARGS)
+{
+	/* Exactly the same as varbit_recv, so share code */
+	return varbit_recv(fcinfo);
+}
+
+/*
+ *		bit_send			- converts bit to binary format
+ */
+Datum
+bit_send(PG_FUNCTION_ARGS)
+{
+	/* Exactly the same as varbit_send, so share code */
+	return varbit_send(fcinfo);
+}
+
 /* bit()
  * Converts a bit() type to a specific internal length.
  * len is the bitlength specified in the column definition.
@@ -228,13 +258,14 @@ bit(PG_FUNCTION_ARGS)
 		PG_RETURN_VARBIT_P(arg);
 
 	if (!isExplicit)
-		elog(ERROR, "Bit string length %d does not match type BIT(%d)",
-			 VARBITLEN(arg), len);
+		ereport(ERROR,
+				(errcode(ERRCODE_STRING_DATA_LENGTH_MISMATCH),
+			   errmsg("bit string length %d does not match type bit(%d)",
+					  VARBITLEN(arg), len)));
 
 	rlen = VARBITTOTALLEN(len);
-	result = (VarBit *) palloc(rlen);
 	/* set to 0 so that string is zero-padded */
-	MemSet(result, 0, rlen);
+	result = (VarBit *) palloc0(rlen);
 	VARATT_SIZEP(result) = rlen;
 	VARBITLEN(result) = len;
 
@@ -312,13 +343,14 @@ varbit_in(PG_FUNCTION_ARGS)
 	if (atttypmod <= 0)
 		atttypmod = bitlen;
 	else if (bitlen > atttypmod)
-		elog(ERROR, "Bit string too long for type BIT VARYING(%d)",
-			 atttypmod);
+		ereport(ERROR,
+				(errcode(ERRCODE_STRING_DATA_LENGTH_MISMATCH),
+				 errmsg("bit string too long for type bit varying(%d)",
+						atttypmod)));
 
 	len = VARBITTOTALLEN(bitlen);
-	result = (VarBit *) palloc(len);
 	/* set to 0 so that *r is always initialised and string is zero-padded */
-	MemSet(result, 0, len);
+	result = (VarBit *) palloc0(len);
 	VARATT_SIZEP(result) = len;
 	VARBITLEN(result) = Min(bitlen, atttypmod);
 
@@ -333,7 +365,11 @@ varbit_in(PG_FUNCTION_ARGS)
 			if (*sp == '1')
 				*r |= x;
 			else if (*sp != '0')
-				elog(ERROR, "Cannot parse '%c' as a binary digit", *sp);
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
+						 errmsg("\"%c\" is not a valid binary digit",
+								*sp)));
+
 			x >>= 1;
 			if (x == 0)
 			{
@@ -354,7 +390,11 @@ varbit_in(PG_FUNCTION_ARGS)
 			else if (*sp >= 'a' && *sp <= 'f')
 				x = (bits8) (*sp - 'a') + 10;
 			else
-				elog(ERROR, "Cannot parse '%c' as a hex digit", *sp);
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
+						 errmsg("\"%c\" is not a valid hexadecimal digit",
+								*sp)));
+
 			if (bc)
 			{
 				*r++ |= x;
@@ -390,9 +430,8 @@ varbit_out(PG_FUNCTION_ARGS)
 	result = (char *) palloc(len + 1);
 	sp = VARBITS(s);
 	r = result;
-	for (i = 0; i <= len - BITS_PER_BYTE; i += BITS_PER_BYTE, sp++)
+	for (i = 0; i < len - BITS_PER_BYTE; i += BITS_PER_BYTE, sp++)
 	{
-		/* print full bytes */
 		x = *sp;
 		for (k = 0; k < BITS_PER_BYTE; k++)
 		{
@@ -400,21 +439,71 @@ varbit_out(PG_FUNCTION_ARGS)
 			x <<= 1;
 		}
 	}
-	if (i < len)
+	x = *sp;
+	for (k = i; k < len; k++)
 	{
-		/* print the last partial byte */
-		x = *sp;
-		for (k = i; k < len; k++)
-		{
-			*r++ = (x & BITHIGH) ? '1' : '0';
-			x <<= 1;
-		}
+		*r++ = (x & BITHIGH) ? '1' : '0';
+		x <<= 1;
 	}
 	*r = '\0';
 
 	PG_RETURN_CSTRING(result);
 }
 
+/*
+ *		varbit_recv			- converts external binary format to varbit
+ *
+ * External format is the bitlen as an int32, then the byte array.
+ */
+Datum
+varbit_recv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+	VarBit	   *result;
+	int			len,
+				bitlen;
+	int			ipad;
+	bits8		mask;
+
+	bitlen = pq_getmsgint(buf, sizeof(int32));
+	if (bitlen < 0)
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_BINARY_REPRESENTATION),
+				 errmsg("invalid length in external bit string")));
+
+	len = VARBITTOTALLEN(bitlen);
+	result = (VarBit *) palloc(len);
+	VARATT_SIZEP(result) = len;
+	VARBITLEN(result) = bitlen;
+
+	pq_copymsgbytes(buf, (char *) VARBITS(result), VARBITBYTES(result));
+
+	/* Make sure last byte is zero-padded if needed */
+	ipad = VARBITPAD(result);
+	if (ipad > 0)
+	{
+		mask = BITMASK << ipad;
+		*(VARBITS(result) + VARBITBYTES(result) - 1) &= mask;
+	}
+
+	PG_RETURN_VARBIT_P(result);
+}
+
+/*
+ *		varbit_send			- converts varbit to binary format
+ */
+Datum
+varbit_send(PG_FUNCTION_ARGS)
+{
+	VarBit	   *s = PG_GETARG_VARBIT_P(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+	pq_sendint(&buf, VARBITLEN(s), sizeof(int32));
+	pq_sendbytes(&buf, VARBITS(s), VARBITBYTES(s));
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
 /* varbit()
  * Converts a varbit() type to a specific internal length.
  * len is the maximum bitlength specified in the column definition.
@@ -438,7 +527,10 @@ varbit(PG_FUNCTION_ARGS)
 		PG_RETURN_VARBIT_P(arg);
 
 	if (!isExplicit)
-		elog(ERROR, "Bit string too long for type BIT VARYING(%d)", len);
+		ereport(ERROR,
+				(errcode(ERRCODE_STRING_DATA_LENGTH_MISMATCH),
+				 errmsg("bit string too long for type bit varying(%d)",
+						len)));
 
 	rlen = VARBITTOTALLEN(len);
 	result = (VarBit *) palloc(rlen);
@@ -808,7 +900,10 @@ bitand(PG_FUNCTION_ARGS)
 	bitlen1 = VARBITLEN(arg1);
 	bitlen2 = VARBITLEN(arg2);
 	if (bitlen1 != bitlen2)
-		elog(ERROR, "Cannot AND bit strings of different sizes");
+		ereport(ERROR,
+				(errcode(ERRCODE_STRING_DATA_LENGTH_MISMATCH),
+				 errmsg("cannot AND bit strings of different sizes")));
+
 	len = VARSIZE(arg1);
 	result = (VarBit *) palloc(len);
 	VARATT_SIZEP(result) = len;
@@ -846,7 +941,9 @@ bitor(PG_FUNCTION_ARGS)
 	bitlen1 = VARBITLEN(arg1);
 	bitlen2 = VARBITLEN(arg2);
 	if (bitlen1 != bitlen2)
-		elog(ERROR, "Cannot OR bit strings of different sizes");
+		ereport(ERROR,
+				(errcode(ERRCODE_STRING_DATA_LENGTH_MISMATCH),
+				 errmsg("cannot OR bit strings of different sizes")));
 	len = VARSIZE(arg1);
 	result = (VarBit *) palloc(len);
 	VARATT_SIZEP(result) = len;
@@ -890,7 +987,10 @@ bitxor(PG_FUNCTION_ARGS)
 	bitlen1 = VARBITLEN(arg1);
 	bitlen2 = VARBITLEN(arg2);
 	if (bitlen1 != bitlen2)
-		elog(ERROR, "Cannot XOR bit strings of different sizes");
+		ereport(ERROR,
+				(errcode(ERRCODE_STRING_DATA_LENGTH_MISMATCH),
+				 errmsg("cannot XOR bit strings of different sizes")));
+
 	len = VARSIZE(arg1);
 	result = (VarBit *) palloc(len);
 	VARATT_SIZEP(result) = len;
@@ -1105,7 +1205,10 @@ bittoint4(PG_FUNCTION_ARGS)
 
 	/* Check that the bit string is not too long */
 	if (VARBITLEN(arg) > sizeof(int4) * BITS_PER_BYTE)
-		elog(ERROR, "Bit string is too large to fit in type integer");
+		ereport(ERROR,
+				(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
+				 errmsg("integer out of range")));
+
 	result = 0;
 	for (r = VARBITS(arg); r < VARBITEND(arg); r++)
 	{
@@ -1149,7 +1252,10 @@ bitfromint8(PG_FUNCTION_ARGS)
 
 	PG_RETURN_VARBIT_P(result);
 #else
-	elog(ERROR, "INT64 is not supported on this platform");
+	ereport(ERROR,
+			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			 errmsg("64-bit integers not supported on this platform")));
+
 	PG_RETURN_NULL();
 #endif
 }
@@ -1164,7 +1270,10 @@ bittoint8(PG_FUNCTION_ARGS)
 
 	/* Check that the bit string is not too long */
 	if (VARBITLEN(arg) > sizeof(result) * BITS_PER_BYTE)
-		elog(ERROR, "Bit string is too large to fit in type int64");
+		ereport(ERROR,
+				(errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
+				 errmsg("integer out of range")));
+
 	result = 0;
 	for (r = VARBITS(arg); r < VARBITEND(arg); r++)
 	{
@@ -1176,7 +1285,10 @@ bittoint8(PG_FUNCTION_ARGS)
 
 	PG_RETURN_INT64(result);
 #else
-	elog(ERROR, "INT64 is not supported on this platform");
+	ereport(ERROR,
+			(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			 errmsg("64-bit integers not supported on this platform")));
+
 	PG_RETURN_NULL();
 #endif
 }
@@ -1256,7 +1368,7 @@ bitposition(PG_FUNCTION_ARGS)
 					mask2 = end_mask << (BITS_PER_BYTE - is);
 					is_match = mask2 == 0;
 #if 0
-					elog(DEBUG3, "S. %d %d em=%2x sm=%2x r=%d",
+					elog(DEBUG4, "S. %d %d em=%2x sm=%2x r=%d",
 						 i, is, end_mask, mask2, is_match);
 #endif
 					break;
