@@ -4,9 +4,9 @@
  * A crippled id utility for use in various shell scripts in use by PostgreSQL
  * (in particular initdb)
  *
- * Copyright (C) 2000 by PostgreSQL Global Development Group
+ * Copyright (c) 2000-2003, PostgreSQL Global Development Group
  *
- * $Header: /cvsroot/pgsql-server/src/bin/pg_id/pg_id.c,v 1.20 2001/03/22 04:00:16 momjian Exp $
+ * $Header: /cvsroot/pgsql-server/src/bin/pg_id/Attic/pg_id.c,v 1.25 2003/10/10 01:34:51 momjian Exp $
  */
 #include "postgres_fe.h"
 
@@ -24,50 +24,70 @@ int
 main(int argc, char *argv[])
 {
 	int			c;
-	int			nameflag = 0,
-				realflag = 0,
-				userflag = 0;
+	int			name_only_flag = 0,
+				use_real_uid_flag = 0,
+				limit_user_info = 0;
 	const char *username = NULL;
-
-	struct passwd *pw;
-
 	extern int	optind;
-	extern char *optarg;
+#ifndef WIN32
+	struct passwd *pw;
+#else
+	struct passwd_win32
+	{
+		int pw_uid;
+		char pw_name[128];
+	} pass_win32;
+	struct passwd_win32 *pw = &pass_win32;
+	unsigned long pwname_size = sizeof(pass_win32.pw_name) - 1;
+
+	pw->pw_uid = 1;
+#endif
 
 	while ((c = getopt(argc, argv, "nru")) != -1)
 	{
 		switch (c)
 		{
 			case 'n':
-				nameflag = 1;
+				name_only_flag = 1;
 				break;
 			case 'r':
-				realflag = 1;
+				use_real_uid_flag = 1;
 				break;
 			case 'u':
-				userflag = 1;
+				limit_user_info = 1;
 				break;
 			default:
+#ifndef WIN32
 				fprintf(stderr, "Usage: %s [-n] [-r] [-u] [username]\n", argv[0]);
+#else
+				fprintf(stderr, "Usage: %s [-n] [-r] [-u]\n", argv[0]);
+#endif
 				exit(1);
 		}
 	}
 
 	if (argc - optind >= 1)
+#ifndef WIN32
 		username = argv[optind];
+#else
+	{
+		fprintf(stderr, "%s: specifying a username is not supported on this platform\n", argv[0]);
+		exit(1);
+	}	
+#endif
 
-	if (nameflag && !userflag)
+	if (name_only_flag && !limit_user_info)
 	{
 		fprintf(stderr, "%s: -n must be used together with -u\n", argv[0]);
 		exit(1);
 	}
-	if (username && realflag)
+	if (username && use_real_uid_flag)
 	{
 		fprintf(stderr, "%s: -r cannot be used when a user name is given\n", argv[0]);
 		exit(1);
 	}
 
-
+#ifndef WIN32
 	if (username)
 	{
 		pw = getpwnam(username);
@@ -77,7 +97,7 @@ main(int argc, char *argv[])
 			exit(1);
 		}
 	}
-	else if (realflag)
+	else if (use_real_uid_flag)
 		pw = getpwuid(getuid());
 	else
 		pw = getpwuid(geteuid());
@@ -87,10 +107,19 @@ main(int argc, char *argv[])
 		perror(argv[0]);
 		exit(1);
 	}
+#else
+	if (!use_real_uid_flag)
+	{
+		fprintf(stderr, "%s: -r must be used on this platform\n", argv[0]);
+		exit(1);
+	}
 
-	if (!userflag)
+	GetUserName(pw->pw_name, &pwname_size);
+#endif
+
+	if (!limit_user_info)
 		printf("uid=%d(%s)\n", (int) pw->pw_uid, pw->pw_name);
-	else if (nameflag)
+	else if (name_only_flag)
 		puts(pw->pw_name);
 	else
 #ifdef __BEOS__
