@@ -1,9 +1,9 @@
 /*
  * psql - the PostgreSQL interactive terminal
  *
- * Copyright 2000 by PostgreSQL Global Development Group
+ * Copyright (c) 2000-2003, PostgreSQL Global Development Group
  *
- * $Header: /cvsroot/pgsql/src/bin/psql/settings.h,v 1.12 2001/10/28 06:25:58 momjian Exp $
+ * $Header: /cvsroot/pgsql-server/src/bin/psql/settings.h,v 1.15 2003/08/04 23:59:40 tgl Exp $
  */
 #ifndef SETTINGS_H
 #define SETTINGS_H
@@ -48,8 +48,7 @@ typedef struct _psqlSettings
 	char	   *inputfile;		/* for error reporting */
 	unsigned	lineno;			/* also for error reporting */
 
-	bool		issuper;		/* is the current user a superuser? (used
-								 * to form the prompt) */
+	bool		timing;			/* enable timing of all queries */
 } PsqlSettings;
 
 extern PsqlSettings pset;
